# Amethyst - An experimental Minecraft server written in Elixir.
# Copyright (C) 2024 KodiCraft
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

defmodule Amethyst.API.Game do
  @moduledoc """
  This module includes the interface for defining and registering
  a game with Amethyst.
  """

  @doc """
  `instantiate/1` is called when a new instance of your game is created. You may start any additional
  processes you want under the `DynamicSupervisor` 'supervisor' and return a map with the PIDs to those
  processes, known as your references. Other callbacks in your game will receive these references and
  may use them to access and update state.
  """
  @callback instantiate(supervisor :: pid()) :: {:ok, state_refs :: map()} | {:error, reason :: term}
  @doc """
  `login/3` is called when a new player logs into a game instance.
  You may either :accept or :reject the player for whatever reason, avoid
  rejecting the player in your default game as that will disconnect the player.
  The PID received in 'from' will be the one that calls all callbacks
  caused directly by this player, such as their movement or interactions --
  You can expect that PID to never be used by any other player and that this same
  player will not use another PID until they disconnect from the server.

  - 'from' is the PID of the player's connection process.
  - 'player_cfg' is a keyword list containing the configuration passed by the game client
  - 'state_refs' are your references (see `instantiate/1`)
  """
  @callback login(from :: pid(), player_cfg :: keyword(), state_refs :: map()) :: :accept | :reject
  @doc """
  `player_position/3` is called when a player moves. This function is called with the absolute coordinates
  that the player client expects. TODO: Teleport Player API.

  - 'from' is the PID of the player's connection process (see `login/3`).
  - 'x', 'y' and 'z' are the absolute coordinates that the player wants to move to.
  - `state_refs` are your references (see `instantiate/1`)
  """
  @callback player_position(from :: pid(), {x :: float(), y :: float(), z :: float()}, state_refs :: map()) :: :ok
  @doc """
  Whether or not this game instance can be joined by a new player. This should include basic logic such as
  if joining makes sense, for instance
  """
  @callback joinable?(state_refs :: map()) :: boolean()

  def child_spec(mod, refs) do
    %{
      id: __MODULE__,
      start: {__MODULE__, :start_link, [mod, refs]}
    }
  end

  @spec start(Amethyst.API.Game, term()) :: {:ok, pid}
  def start(mod, refs) do
    pid = spawn(fn ->
      true = refs |> Enum.all?(fn {_key, pid } -> Process.link(pid) end)
      loop(mod, refs)
    end)
    {:ok, pid}
  end
  def start_link(mod, refs) do
    {:ok, pid} = start(mod, refs)
    Process.link(pid)
    {:ok, pid}
  end
  defp loop(mod, refs) do
    receive do
      {:login, caller, cfg} ->
        Task.Supervisor.start_child(refs[:task_supervisor],
        fn -> mod.login(caller, cfg, refs) end)
      {:player_position, caller, pos} ->
        Task.Supervisor.start_child(refs[:task_supervisor],
        fn -> mod.player_position(caller, pos, refs) end)
    end
    loop(mod, refs)
  end

  def login(pid, cfg) do
    send(pid, {:login, self(), cfg})
    receive do
      :ok -> :ok
      :refuse -> :refuse
    end
  end

  @spec player_position(atom() | pid() | port() | reference() | {atom(), atom()}, {float(), float(), float()}) :: {float(), float(), float()}
  def player_position(pid, pos) do
    send(pid, {:player_position, self(), pos})
    receive do
      pos -> pos
    end
  end
end
